from datetime import timedelta
from os import environ
from typing import Any, Union, Tuple
from redis import Redis
from .stateful_lambda import AbstractCacheHolder
from .utils import parse

_Time = Union[None, int, timedelta]
_Value = Union[bytes, float, int, str]
_Key = Union[str, bytes]


class ExtendRedis(Redis, AbstractCacheHolder):
    def setnx_ttl(
        self,
        name: _Key,
        value: _Value,
        ex: _Time = None,
        px: _Time = None,
        keepttl: bool = None,
        exat: Any = None,
        pxat: Any = None,
    ) -> bool:
        """Set if Not Exists with specified expire time

            Args:
                name (_Key): key
                value (_Value): value
                ex (_Time, optional): seconds -- Set the specified expire time, in seconds. Defaults to None.
                px (_Time, optional): milliseconds -- Set the specified expire time, in milliseconds. Defaults to None
                keepttl (bool, optional): Retain the time to live associated with the key. Defaults to None
                exat (Any, optional): timestamp-seconds -- Set the specified Unix time at which the key will expire, in seconds. Defaults to None
                pxat (Any, optional): timestamp-milliseconds -- Set the specified Unix time at which the key will expire, in milliseconds. Defaults to None

            Returns:
                Pipeline[_StrType]: True if Key is not exists, otherwise False.
            """
        return self.set(
            name=name,
            value=value,
            nx=True,
            ex=ex,
            px=px,
            keepttl=keepttl,
            exat=exat,
            pxat=pxat,
        )

    def setnx_dedup(
        self,
        name: _Key,
        value: _Value,
        ex: _Time = None,
        px: _Time = None,
        keepttl: bool = None,
        exat: Any = None,
        pxat: Any = None,
    ) -> Tuple[bool, _Value]:
        """Set if Not Exists and Not Duplicated Value

            Args:
                name (_Key): key
                value (_Value): value
                ex (_Time, optional): seconds -- Set the specified expire time, in seconds. Defaults to None.
                px (_Time, optional): milliseconds -- Set the specified expire time, in milliseconds. Defaults to None
                keepttl (bool, optional): Retain the time to live associated with the key. Defaults to None
                exat (Any, optional): timestamp-seconds -- Set the specified Unix time at which the key will expire, in seconds. Defaults to None
                pxat (Any, optional): timestamp-milliseconds -- Set the specified Unix time at which the key will expire, in milliseconds. Defaults to None

            Returns:
                Pipeline[_StrType]: True if Key is not exists, or new value is same to old value.
            """
        if self.set(
            name=name,
            value=value,
            nx=True,
            ex=ex,
            px=px,
            keepttl=keepttl,
            exat=exat,
            pxat=pxat,
        ):
            return [True, None]

        val = self.get(name=name)
        if val == value:
            return [True, value]
        return [False, val]

    @classmethod
    def from_environ(cls, **kw):
        return cls(
            host=environ.get("REDIS_HOST", "localhost"),
            port=parse(int, environ.get("REDIS_PORT"), 6379),
            db=parse(int, environ.get("REDIS_DB"), 0),
            password=environ.get("REDIS_PASSWORD"),
            username=environ.get("REDIS_USER"),
            **kw
        )
