from typing import Any, Dict, Iterator, Tuple
from datetime import datetime

MSG = 'msg'
TOPIC = 'topic'
SENT_TIME = 'sentTime'  # when message sent
ARRIVE_PIPE = 'arrivePipe'  # when message arrive at pipe (Queue, Pub/Sub, Stream broker)
EVENT_SOURCE = 'eventSource'


def parse_timestamp_ms(timestamp_ms):
    try:
        return datetime.fromtimestamp(float(timestamp_ms) / 1000)
    except:
        return None


'''
{
  "Records": [
    {
      "messageId": "19dd0b57-b21e-4ac1-bd88-01bbb068cb78",
      "receiptHandle": "MessageReceiptHandle",
      "body": "Hello from SQS!",
      "attributes": {
        "ApproximateReceiveCount": "1",
        "SentTimestamp": "1523232000000",
        "SenderId": "123456789012",
        "ApproximateFirstReceiveTimestamp": "1523232000001"
      },
      "messageAttributes": {},
      "md5OfBody": "{{{md5_of_body}}}",
      "eventSource": "aws:sqs",
      "eventSourceARN": "arn:aws:sqs:us-east-1:123456789012:MyQueue",
      "awsRegion": "us-east-1"
    }
  ]
}
'''


def SQS(record: Dict[str, Any]) -> Dict[str, Any]:
    attr = record.get('attributes', {})

    return {
        MSG:
        record.get('body'),
        SENT_TIME:
        parse_timestamp_ms(attr.get('SentTimestamp')),
        ARRIVE_PIPE:
        parse_timestamp_ms(attr.get('ApproximateFirstReceiveTimestamp'))
    }


'''
{
  "Records": [
    {
      "EventSource": "aws:sns",
      "EventVersion": "1.0",
      "EventSubscriptionArn": "arn:aws:sns:us-east-1:{{{accountId}}}:ExampleTopic",
      "Sns": {
        "Type": "Notification",
        "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
        "TopicArn": "arn:aws:sns:us-east-1:123456789012:ExampleTopic",
        "Subject": "example subject",
        "Message": "example message",
        "Timestamp": "1970-01-01T00:00:00.000Z",
        "SignatureVersion": "1",
        "Signature": "EXAMPLE",
        "SigningCertUrl": "EXAMPLE",
        "UnsubscribeUrl": "EXAMPLE",
        "MessageAttributes": {
          "Test": {
            "Type": "String",
            "Value": "TestString"
          },
          "TestBinary": {
            "Type": "Binary",
            "Value": "TestBinary"
          }
        }
      }
    }
  ]
}
'''


def SNS(record: Dict[str, Any]) -> Dict[str, Any]:
    attr = record.get('attributes', {})

    return {
        MSG: record.get('Message'),
        TOPIC: record.get('Subject'),
        SENT_TIME: parse_timestamp_ms(attr.get('Timestamp')),
        ARRIVE_PIPE: None
    }


def default_parse(record):
    return {'msg': record}


CASE = {'aws:sqs': SQS, 'aws:sns': SNS, None: default_parse}


def parse_record(records) -> Tuple[int, Iterator[Dict[str, Any]]]:
    if 'Records' in records:
        # AWS product
        def gen():
            for record in records['Records']:
                source = record.get(EVENT_SOURCE)
                yield {
                    EVENT_SOURCE: source,
                    'record': record,
                    **CASE.get(source, default_parse)(record)
                }

        return [len(records['Records']), gen]

    elif 'records' in records:
        # Kafka, MSK, Kinesis Analytics, MQ
        source = records.get(EVENT_SOURCE)
        parser = CASE.get(source, default_parse)

        def gen():
            for record in records['records']:
                yield {
                    EVENT_SOURCE: source,
                    'record': record,
                    **parser(record)
                }

        return [len(records['records']), gen]
    return [0, None]
