from datetime import timedelta
from typing import Any, Callable, Dict, Optional, Union, Tuple


class AbstractCacheHolder:
    def setnx_ttl(
        self, name: str, value: Any, ex: Optional[float] = None, *args, **kw
    ) -> bool:
        """Set if Not Exists with specified expire time

        Args:
            name (_Key): key
            value (_Value): value
            ex (_Time, optional): seconds -- Set the specified expire time, in seconds. Defaults to None.
            px (_Time, optional): milliseconds -- Set the specified expire time, in milliseconds. Defaults to None
            keepttl (bool, optional): Retain the time to live associated with the key. Defaults to None
            exat (Any, optional): timestamp-seconds -- Set the specified Unix time at which the key will expire, in seconds. Defaults to None
            pxat (Any, optional): timestamp-milliseconds -- Set the specified Unix time at which the key will expire, in milliseconds. Defaults to None

        Returns:
            Pipeline[_StrType]: True if Key is not exists, otherwise False.
        """
        raise NotImplementedError()

    def setnx_dedup(
        self, name: str, value: Any, ex: Optional[float] = None, *args, **kw
    ) -> Tuple[bool, Any]:
        """Set if Not Exists and Not Duplicated Value

        Args:
            name (_Key): key
            value (_Value): value
            ex (_Time, optional): seconds -- Set the specified expire time, in seconds. Defaults to None.
            px (_Time, optional): milliseconds -- Set the specified expire time, in milliseconds. Defaults to None
            keepttl (bool, optional): Retain the time to live associated with the key. Defaults to None
            exat (Any, optional): timestamp-seconds -- Set the specified Unix time at which the key will expire, in seconds. Defaults to None
            pxat (Any, optional): timestamp-milliseconds -- Set the specified Unix time at which the key will expire, in milliseconds. Defaults to None

        Returns:
            Pipeline[_StrType]: True if Key is not exists, or new value is same to old value.
        """
        raise NotImplementedError()

    def exists(self, name: str, *args, **kw) -> bool:
        """Check Whether key Exists 

        Args:
            name (_Key): key
        Returns:
            bool: True if Key is not exists, and new value is not same to old value.
        """
        raise NotImplementedError()

    def get(self, name: str) -> Any:
        raise NotImplementedError()

    def set(self, name: str, value: Any, *args, **kw):
        raise NotImplementedError()


class StatefulClient:
    def __init__(
        self,
        cache_holder: AbstractCacheHolder,
        re_invoke_callback: Optional[Callable[[Union[str, bytes]], Any]] = None,
    ) -> None:
        self.holder = cache_holder
        self.reinvoke = re_invoke_callback

    def get_state_data(
        self,
        name: str,
        hold_nx: bool,
        inputs: Any,
        ttl_s: Union[float, int, timedelta, None] = None,
    ) -> Tuple[bool, Any]:
        """Get state data.
        If there is state: return [True, `state`]
        Else if hold_nx is True: set the `inputs` into cache and return [False, `cache response`]
        Else: Put input into queue and return [False, `queue response`]

        Args:
            name (str): Name of state
            hold_nx (bool): Hold the input if there is not state
            inputs (Any): Input data to be set as state if hold_nx is True.
            ttl_s (Union[float, int, timedelta, None], optional): [description]. Defaults to None.

        Returns:
            Tuple[bool, Any]: 1st value is True whether there is state. Another is state or cache response or queue response.
        """
        if hold_nx:
            cache, val = self.holder.setnx_dedup(
                name=name, value=inputs, ex=s_to_timedelta(ttl_s)
            )
            if cache:
                # Can set = No state
                return [False, val]
            # there is state
            return [True, val]
        elif self.holder.exists(name=name):
            return [True, self.holder.get(name=name)]
        else:
            return [False, self.reinvoke(inputs)]

    def set_state_nx(
        self, name: str, value: Any, ttl_s: Union[float, int, timedelta, None] = None
    ) -> Tuple[bool, Any]:
        """Set state if not exists

        Args:
            name (str): Key
            value (Any): value
            ttl_s (Union[float, int, timedelta, None], optional): [description]. Defaults to None.

        Returns:
            (Tuple[bool, Any]): True whether can set state and it don't exist. And invoke response
        """

        if self.holder.setnx_ttl(name=name, value=value, ex=s_to_timedelta(ttl_s)):
            return [True, None]
        else:
            return [False, self.reinvoke(value)]

    def set(
        self, name: str, value: Any, ttl_s: Union[float, int, timedelta, None] = None
    ) -> bool:
        """Set state

        Args:
            name (str): Key
            value (Any): value
            ttl_s (Union[float, int, timedelta, None], optional): [description]. Defaults to None.

        Returns:
            (Tuple[bool, Any]): True whether can set state and it don't exist. And invoke response
        """
        return self.holder.set(name=name, value=value, ex=s_to_timedelta(ttl_s))

    def reinvoke(self, inputs: Any):
        return None


def s_to_timedelta(s: Union[float, int, timedelta, None]):
    if s is None:
        return None
    if isinstance(s, timedelta):
        return s
    return timedelta(seconds=s)


def get_lambda_fn(context: Dict[str, Any]) -> str:
    return context.get("invoked_function_arn") or context.get("function_name")

